# coding=utf-8
# Packages streamsx depends on and thus must be installed
# These packages are not added into the Streams application.

_DEP_PACKAGES = [
    "dill",
    "future",
    "enum34",
    "requests"
 ]
